@@ -11,21 +11,23 @@
  */
 #include <config.h>
 #include <stdarg.h>
-#include <unistd.h>
 #include <fcntl.h>
 #include <errno.h>
 #include <string.h>
 #include <stdio.h>
+#include <ctype.h>
 
 #ifdef LINK_SSL_SUPPORT
-#include <openssl/ssl.h>
+#  include <openssl/ssl.h>
 #endif
 
 #include "linc-private.h"
+#include "linc-compat.h"
 #include <linc/linc-config.h>
 #include <linc/linc-connection.h>
 
 static GObjectClass *parent_class = NULL;
+static guint _link_timeout = 0;
 
 enum {
 	BROKEN,
@@ -47,13 +49,17 @@ static gboolean link_connection_io_handler (GIOChannel  *gioc,
 					    GIOCondition condition,
 					    gpointer     data);
 
-#define link_connection_ref_T(cnx) g_object_ref (cnx)
+static inline LinkConnection*
+link_connection_ref_T (LinkConnection *cnx) 
+{
+	return LINK_CONNECTION (g_object_ref (G_OBJECT (cnx)));
+}
 
-gpointer
-link_connection_ref (gpointer cnx)
+LinkConnection *
+link_connection_ref (LinkConnection *cnx)
 {
 	CNX_AND_LIST_LOCK (cnx);
-	g_object_ref (cnx);
+	g_object_ref (G_OBJECT (cnx));
 	CNX_AND_LIST_UNLOCK (cnx);
 
 	return cnx;
@@ -61,19 +67,19 @@ link_connection_ref (gpointer cnx)
 
 /* Only call if we are _certain_ that we don't hold the last ref */
 static void
-link_connection_unref_T_ (gpointer cnx)
+link_connection_unref_T_ (LinkConnection *cnx)
 {
 	g_assert (((GObject *)cnx)->ref_count > 1);
-	g_object_unref (cnx);
+	g_object_unref (G_OBJECT (cnx));
 }
 
 static void
-link_connection_unref_unlock (gpointer cnx)
+link_connection_unref_unlock (LinkConnection *cnx)
 {
 	gboolean tail_unref = FALSE;
 
 	if (((GObject *)cnx)->ref_count > 1)
-		g_object_unref (cnx);
+		g_object_unref (G_OBJECT (cnx));
 
 	else {
 		cnx_list = g_list_remove (cnx_list, cnx);
@@ -95,10 +101,10 @@ link_connection_unref_unlock (gpointer cnx)
 void
 link_connection_exec_cnx_unref (LinkCommandCnxUnref *cmd, gboolean immediate)
 {
-	d_printf ("Exec defered unref on %p", cmd->cnx);
+	d_printf ("Exec defered unref on %p\n", cmd->cnx);
 
 	if (immediate) /* In I/O thread - with just 1 ref left */
-		g_object_unref (cmd->cnx);
+		g_object_unref (G_OBJECT (cmd->cnx));
 	else {
 		CNX_AND_LIST_LOCK (cmd->cnx);
 		link_connection_unref_unlock (cmd->cnx);
@@ -106,7 +112,7 @@ link_connection_exec_cnx_unref (LinkCommandCnxUnref *cmd, gboolean immediate)
 }
 
 void
-link_connection_unref (gpointer cnx)
+link_connection_unref (LinkConnection *cnx)
 {
 	g_return_if_fail (cnx != NULL);
 
@@ -119,9 +125,8 @@ static void
 link_close_fd (LinkConnection *cnx)
 {
 	if (cnx->priv->fd >= 0) {
-		d_printf ("Close %d\n", cnx->priv->fd);
-
-		LINK_CLOSE (cnx->priv->fd);
+		d_printf ("link_close_fd: closing %d\n", cnx->priv->fd);
+		LINK_CLOSE_SOCKET (cnx->priv->fd);
 	}
 	cnx->priv->fd = -1;
 }
@@ -144,28 +149,66 @@ link_connection_emit_broken (LinkConnection *cnx, GSList *callbacks)
 	g_slist_free (callbacks);
 }
 
+/*
+ * Unfortunate to have a global list, but we need to know
+ * if this is being processed in the main thread & if so
+ * simply append to it.
+ */
+static GSList *idle_broken_cnxs = NULL;
+
 static gboolean
-link_connection_broken_idle (gpointer data)
+link_connection_broken_idle (gpointer dummy)
 {
 	GSList *callbacks;
-	LinkConnection *cnx = data;
+	LinkConnection *cnx;
 
-	d_printf ("Connection %p broken idle ...\n", data);
+	d_printf ("Connection broken idle ...\n");
 
-	CNX_LOCK (cnx);
-	callbacks = cnx->idle_broken_callbacks;
-	cnx->idle_broken_callbacks = NULL;
-	cnx->inhibit_reconnect = FALSE;
-	link_signal ();
-	CNX_UNLOCK (cnx);
+	do {
+		link_lock();
+		cnx = NULL;
+		if (idle_broken_cnxs != NULL) {
+			cnx = idle_broken_cnxs->data;
+			idle_broken_cnxs = g_slist_delete_link (idle_broken_cnxs, idle_broken_cnxs);
+		}
+		if (cnx) {
+			callbacks = cnx->idle_broken_callbacks;
+			cnx->idle_broken_callbacks = NULL;
+			cnx->inhibit_reconnect = FALSE;
+			link_signal ();
+		}
+		link_unlock ();
 
-	link_connection_emit_broken (cnx, callbacks);
-
-	link_connection_unref (cnx);
+		if (cnx) {
+			link_connection_emit_broken (cnx, callbacks);
+			link_connection_unref (cnx);
+		}
+	} while (cnx != NULL);
 
 	return FALSE;
 }
 
+static void
+add_idle_broken_for_cnx_T (LinkConnection *cnx)
+{
+	if (idle_broken_cnxs != NULL) {
+		fprintf (stderr, "Deadlock potential - avoiding evil bug!\n");
+		/*
+		 * just append ourself & exit - we don't want to
+		 * inhibit re-connection because of the horrendous
+		 * deadlock possible, cf. g#534351#
+		 */
+		if (g_slist_find (idle_broken_cnxs, cnx) != NULL)
+			return;
+	} else {
+		/* inhibit reconnect while we emit 'broken' */
+		cnx->inhibit_reconnect = TRUE;
+		g_idle_add (link_connection_broken_idle, NULL);
+	}
+	link_connection_ref_T (cnx);
+	idle_broken_cnxs = g_slist_prepend (idle_broken_cnxs, cnx);
+}
+
 static void
 link_source_remove (LinkConnection *cnx)
 {
@@ -218,6 +261,19 @@ queue_free (LinkConnection *cnx)
 	cnx->priv->write_queue = NULL;
 }
 
+static void
+dispatch_callbacks_drop_lock (LinkConnection *cnx)
+{
+	GSList *callbacks;
+
+	callbacks = cnx->idle_broken_callbacks;
+	cnx->idle_broken_callbacks = NULL;
+
+	CNX_UNLOCK (cnx);
+	link_connection_emit_broken (cnx, callbacks);
+	CNX_LOCK (cnx);
+}
+
 /*
  * link_connection_class_state_changed:
  * @cnx: a #LinkConnection
@@ -275,6 +331,7 @@ link_connection_state_changed_T_R (LinkConnection      *cnx,
 		break;
 
 	case LINK_DISCONNECTED:
+	case LINK_TIMEOUT:
 		link_source_remove (cnx);
 		link_close_fd (cnx);
 		queue_free (cnx);
@@ -290,22 +347,12 @@ link_connection_state_changed_T_R (LinkConnection      *cnx,
 
 			if (cnx->idle_broken_callbacks) {
 				if (!link_thread_io ()) {
-					GSList *callbacks;
-
 					d_printf ("Immediate broken callbacks at immediately\n");
-
-					callbacks = cnx->idle_broken_callbacks;
-					cnx->idle_broken_callbacks = NULL;
-
-					CNX_UNLOCK (cnx);
-					link_connection_emit_broken (cnx, callbacks);
-					CNX_LOCK (cnx);
+				
+					dispatch_callbacks_drop_lock (cnx);
 				} else {
 					d_printf ("Queuing broken callbacks at idle\n");
-
-					cnx->inhibit_reconnect = TRUE;
-					link_connection_ref_T (cnx);
-					g_idle_add (link_connection_broken_idle, cnx);
+					add_idle_broken_for_cnx_T (cnx);
 				}
 			}
 		}
@@ -434,6 +481,18 @@ link_connection_from_fd_T (LinkConnection         *cnx,
 	g_free (cnx->remote_serv_info);
 	cnx->remote_serv_info = remote_serv_info;
 
+	switch (cnx->proto->family) {
+	case AF_INET:
+#ifdef AF_INET6
+	case AF_INET6:
+#endif
+		if (_link_timeout && !cnx->timeout_msec) /* this should'nt happen twice but I'm always paranoid... */
+			cnx->timeout_msec = _link_timeout;
+		break;
+	default:
+		break;
+	}
+
 	d_printf ("Cnx from fd (%d) '%s', '%s', '%s'\n",
 		 fd, proto->name, 
 		 remote_host_info ? remote_host_info : "<Null>",
@@ -489,6 +548,25 @@ link_connection_from_fd (LinkConnection         *cnx,
 	CNX_UNLOCK (cnx);
 }
 
+#ifndef G_OS_WIN32
+static void
+fix_permissions (const char *filename)
+{
+	char *tmp_dir = g_strdup (filename);
+	char *p;
+	struct stat stat_buf;
+
+	if (!tmp_dir)
+		return;
+	p = strrchr (tmp_dir, '/');
+	if (p) {
+		*p = '\0';
+		stat (tmp_dir, &stat_buf);
+		chown (filename, stat_buf.st_uid, -1);
+	}
+}
+#endif
+
 static gboolean
 link_connection_do_initiate (LinkConnection        *cnx,
 			     const char            *proto_name,
@@ -523,23 +601,94 @@ link_connection_do_initiate (LinkConnection        *cnx,
 
 	fd = socket (proto->family, SOCK_STREAM, 
 		     proto->stream_proto_num);
+#ifdef HAVE_WINSOCK2_H
+	if (fd == INVALID_SOCKET) {
+		fd = -1;
+		link_map_winsock_error_to_errno ();
+	}
+#endif
 
-	if (fd < 0)
+	if (fd < 0) {
+		d_printf ("socket() failed: %s\n", link_strerror (errno));
 		goto out;
+	}
 
-	if (options & LINK_CONNECTION_NONBLOCKING)
+	if (options & LINK_CONNECTION_NONBLOCKING) {
+#ifdef HAVE_WINSOCK2_H
+		u_long yes = 1;
+		if (ioctlsocket (fd, FIONBIO, &yes) == SOCKET_ERROR) {
+			link_map_winsock_error_to_errno ();
+			d_printf ("ioctlsocket(FIONBIO) failed: %s\n",
+				  link_strerror (errno));
+			goto out;
+		}
+#else
 		if (fcntl (fd, F_SETFL, O_NONBLOCK) < 0)
 			goto out;
+#endif
+	}
 
+#if defined (F_SETFD) && defined (FD_CLOEXEC)
 	if (fcntl (fd, F_SETFD, FD_CLOEXEC) < 0)
 		goto out;
+#endif
+#ifdef HAVE_WINSOCK2_H
+	{
+		SOCKET newfd;
 
-	rv = connect (fd, saddr, saddr_len);
+		if (!DuplicateHandle (GetCurrentProcess (), (HANDLE) fd,
+				      GetCurrentProcess (), (LPHANDLE) &newfd,
+				      0, FALSE, DUPLICATE_SAME_ACCESS | DUPLICATE_CLOSE_SOURCE)) {
+			d_printf ("DuplicateHandle failed: %s\n", link_strerror (WSAGetLastError ()));
+			return FALSE;
+		}
+		fd = newfd;
+	}
+#endif	
+
+#ifndef G_OS_WIN32
+	if (!strcmp (proto_name, "UNIX") && getuid() == 0) {
+		fix_permissions (service);
+	}
+#endif
+
+	LINK_TEMP_FAILURE_RETRY_SOCKET (connect (fd, saddr, saddr_len), rv);
+#ifdef HAVE_WINSOCK2_H
+	if (rv == SOCKET_ERROR) {
+		if ((options & LINK_CONNECTION_NONBLOCKING) &&
+		    WSAGetLastError () == WSAEWOULDBLOCK) {
+			/* connect() for nonblocking sockets always
+			 * fails with WSAEWOULDBLOCK. We have to
+			 * select() to wait for actual status.
+			 */
+			fd_set write_fds, except_fds;
+
+			FD_ZERO (&write_fds);
+			FD_SET (fd, &write_fds);
+			
+			FD_ZERO (&except_fds);
+			FD_SET (fd, &except_fds);
+			
+			rv  = select (1, NULL, &write_fds, &except_fds, NULL);
+			if (rv == SOCKET_ERROR) {
+				rv = -1;
+				link_map_winsock_error_to_errno ();
+			} else if (FD_ISSET (fd, &write_fds)) {
+				rv = 0;
+			} else if (FD_ISSET (fd, &except_fds)) {
+				rv = -1;
+				errno = WSAECONNREFUSED;
+			}
+		} else {
+			rv = -1;
+			link_map_winsock_error_to_errno ();
+		}
+	}
+#endif
 	if (rv && errno != EINPROGRESS)
 		goto out;
 
-	d_printf ("initiate 'connect' on new fd %d [ %d; %d ]\n",
-		 fd, rv, errno);
+	d_printf ("initiate 'connect' on new fd %d\n", fd);
 
 	g_assert (CNX_IS_LOCKED (0));
 	link_connection_from_fd_T
@@ -551,8 +700,9 @@ link_connection_do_initiate (LinkConnection        *cnx,
 
  out:
 	if (!retval && fd >= 0) {
-		d_printf ("initiation failed\n");
-		LINK_CLOSE (fd);
+		d_printf ("initiation failed: %s\n", link_strerror (errno));
+		d_printf ("closing %d\n", fd);
+		LINK_CLOSE_SOCKET (fd);
 	}
 
 	g_free (saddr);
@@ -581,15 +731,27 @@ link_connection_try_reconnect (LinkConnection *cnx)
 	d_printf ("Try for reconnection on %p: %d\n",
 		  cnx, cnx->inhibit_reconnect);
 
-	while (cnx->inhibit_reconnect)
-		link_wait ();
+	while (cnx->inhibit_reconnect) {
+		if (g_main_context_acquire (NULL)) {
+			d_printf ("Dispatch callbacks in 'main' (mainloop owning) thread\n");
+			cnx->inhibit_reconnect = FALSE;
+			dispatch_callbacks_drop_lock (cnx);
+			g_main_context_release (NULL);
+		} else 
+			link_wait ();
+	}
 
-	if (cnx->status != LINK_DISCONNECTED)
-		g_warning ("trying to re-connect connected cnx.");
-	else
+	switch (cnx->status) {
+	case LINK_DISCONNECTED :
+	case LINK_TIMEOUT :
 		link_connection_do_initiate
 			(cnx, cnx->proto->name, cnx->remote_host_info,
 			 cnx->remote_serv_info, cnx->options);
+		break;
+	default :
+		g_warning ("trying to re-connect connected cnx.");
+		break;
+	}
 
 	cnx->priv->was_disconnected = TRUE;
 	status = link_connection_wait_connected_T (cnx);
@@ -727,8 +889,18 @@ link_connection_read (LinkConnection *cnx,
 			n = SSL_read (cnx->priv->ssl, buf, len);
 		else
 #endif
-			n = read (cnx->priv->fd, buf, len);
-
+#ifdef HAVE_WINSOCK2_H
+			if ((n = recv (cnx->priv->fd, buf, len, 0)) == SOCKET_ERROR) {
+				n = -1;
+				link_map_winsock_error_to_errno ();
+				d_printf ("recv failed: %s\n",
+					  link_strerror (errno));
+			}
+#else
+			LINK_TEMP_FAILURE_RETRY_SYSCALL (read (cnx->priv->fd, 
+							       buf, 
+							       len), n);
+#endif
 		g_assert (n <= len);
 
 		if (n < 0) {
@@ -770,10 +942,16 @@ link_connection_read (LinkConnection *cnx,
 			buf += n;
 			len -= n;
 			bytes_read += n;
+#ifdef CONNECTION_DEBUG
+			cnx->priv->total_read_bytes += n;
+#endif
 		}
 	} while (len > 0 && block_for_full_read);
 
-	d_printf ("we read %d bytes\n", bytes_read);
+#ifdef CONNECTION_DEBUG
+	d_printf ("we read %d bytes (total %"G_GUINT64_FORMAT")\n",
+		  bytes_read, cnx->priv->total_read_bytes);
+#endif
 
  out:
 	CNX_UNLOCK (cnx);
@@ -851,11 +1029,32 @@ write_data_T (LinkConnection *cnx, QueuedWrite *qw)
 				       qw->vecs->iov_len);
 		else
 #endif
-			n = writev (cnx->priv->fd, qw->vecs,
-				    MIN (qw->nvecs, LINK_IOV_MAX));
-
-		d_printf ("wrote %d bytes\n", n);
-
+#ifdef HAVE_WINSOCK2_H
+			{
+				if (WSASend (cnx->priv->fd, qw->vecs,
+					     MIN (qw->nvecs, LINK_IOV_MAX),
+					     (LPDWORD) &n, 0, NULL, NULL) == SOCKET_ERROR) {
+					if (WSAGetLastError () == WSAEWOULDBLOCK)
+						link_win32_watch_set_write_wouldblock (cnx->priv->tag, TRUE);
+					n = -1;
+					link_map_winsock_error_to_errno ();
+					d_printf ("WSASend failed: %s\n",
+						  link_strerror (errno));
+				} else {
+					link_win32_watch_set_write_wouldblock (cnx->priv->tag, FALSE);
+				}
+			}
+#else
+			LINK_TEMP_FAILURE_RETRY_SOCKET (writev (cnx->priv->fd,
+								qw->vecs,
+								MIN (qw->nvecs, LINK_IOV_MAX)), n);
+#endif
+#ifdef CONNECTION_DEBUG
+		d_printf ("wrote %d bytes (total %"G_GUINT64_FORMAT")\n",
+			  n,
+			  (cnx->priv->total_written_bytes += ((n > 0) ? n : 0),
+			   cnx->priv->total_written_bytes));
+#endif
 		if (n < 0) {
 #ifdef LINK_SSL_SUPPORT
 			if (cnx->options & LINK_CONNECTION_SSL) {
@@ -945,7 +1144,7 @@ link_connection_flush_write_queue_T_R (LinkConnection *cnx)
 
 		} else {
 			if (status == LINK_IO_FATAL_ERROR) {
-				d_printf ("Fatal error on queued write");
+				d_printf ("Fatal error on queued write\n");
 				link_connection_state_changed_T_R (cnx, LINK_DISCONNECTED);
 				
 			} else {
@@ -971,7 +1170,7 @@ link_connection_flush_write_queue_T_R (LinkConnection *cnx)
 void
 link_connection_exec_set_condition (LinkCommandSetCondition *cmd, gboolean immediate)
 {
-	d_printf ("Exec defered set condition on %p -> 0x%x",
+	d_printf ("Exec defered set condition on %p -> 0x%x\n",
 		  cmd->cnx, cmd->condition);
 
 	if (!immediate)
@@ -1135,9 +1334,10 @@ link_connection_finalize (GObject *obj)
 
 	g_free (cnx->priv);
 
-#ifdef G_ENABLE_DEBUG
-	g_assert (g_list_find(cnx_list, cnx) == NULL);
-#endif
+	if (cnx->timeout_mutex)
+		g_mutex_free (cnx->timeout_mutex);
+	if (cnx->timeout_source_id) 
+		link_io_thread_remove_timeout (cnx->timeout_source_id);
 
 	parent_class->finalize (obj);
 }
@@ -1150,6 +1350,17 @@ link_connection_init (LinkConnection *cnx)
 	cnx->priv = g_new0 (LinkConnectionPrivate, 1);
 	cnx->priv->fd = -1;
 	cnx->priv->was_disconnected = FALSE;
+
+	cnx->timeout_mutex = NULL;
+	cnx->timeout_msec = 0;
+	cnx->timeout_source_id = 0;
+	cnx->timeout_status = LINK_TIMEOUT_UNKNOWN;
+	cnx->tdata = NULL; 
+
+#ifdef CONNECTION_DEBUG
+	cnx->priv->total_read_bytes = 0;
+	cnx->priv->total_written_bytes = 0;
+#endif
 }
 
 static void
@@ -1196,7 +1407,7 @@ link_connection_get_type (void)
 			NULL,           /* class_data */
 			sizeof (LinkConnection),
 			0,              /* n_preallocs */
-			(GInstanceInitFunc) link_connection_init,
+			(GInstanceInitFunc) link_connection_init
 		};
       
 		object_type = g_type_register_static (G_TYPE_OBJECT,
@@ -1277,7 +1488,7 @@ link_connection_io_handler (GIOChannel  *gioc,
 		switch (cnx->status) {
 		case LINK_CONNECTING:
 			n = 0;
-			rv = getsockopt (cnx->priv->fd, SOL_SOCKET, SO_ERROR, &n, &n_size);
+			rv = getsockopt (cnx->priv->fd, SOL_SOCKET, SO_ERROR, (char *) &n, &n_size);
 			if (!rv && !n && condition == G_IO_OUT) {
 				d_printf ("State changed to connected on %d\n", cnx->priv->fd);
 
@@ -1323,7 +1534,7 @@ link_connection_get_status (LinkConnection *cnx)
 	status = cnx->status;
 	CNX_UNLOCK (cnx);
 
-	d_printf ("Get status on %p = %d", cnx, status);
+	d_printf ("Get status on %p = %d\n", cnx, status);
 
 	return status;
 }
@@ -1331,7 +1542,7 @@ link_connection_get_status (LinkConnection *cnx)
 void
 link_connection_exec_disconnect (LinkCommandDisconnect *cmd, gboolean immediate)
 {
-	d_printf ("Exec defered disconnect on %p", cmd->cnx);
+	d_printf ("Exec defered disconnect on %p\n", cmd->cnx);
 
 	link_connection_state_changed (cmd->cnx, LINK_DISCONNECTED);
 
@@ -1445,3 +1656,10 @@ link_connections_close (void)
 
 	g_list_free (cnx);
 }
+
+void
+link_set_timeout (guint msec)
+{
+	_link_timeout = msec;
+}
+
