@@ -6,6 +6,7 @@
 #include <sys/time.h>
 #include <ctype.h>
 #include <orbit/orbit.h>
+#include <stdlib.h>
 
 #include "orb-core-private.h"
 #include "orbit-debug.h"
@@ -84,6 +85,8 @@ ORBit_trace_value (gconstpointer *val, CORBA_TypeCode tc)
 	_ORBIT_HANDLE_TYPE (enum, long, int, 32, 4, "%d");
 	_ORBIT_HANDLE_TYPE (ushort, unsigned_short, unsigned int, 16, 2, "%u");
 	_ORBIT_HANDLE_TYPE (ulong, unsigned_long, unsigned int, 32, 4, "0x%x");
+	_ORBIT_HANDLE_TYPE (longlong, long_long, long long, 64, 8, "%"G_GINT64_FORMAT);
+	_ORBIT_HANDLE_TYPE (ulonglong, unsigned_long_long, unsigned long long, 64, 8, "%"G_GUINT64_FORMAT);
 	_ORBIT_HANDLE_TYPE (boolean, boolean, int, 8, 1, "%d");
 	_ORBIT_HANDLE_TYPE (char, char, int, 8, 1, "'%c'");
 	_ORBIT_HANDLE_TYPE (wchar, wchar, int, 16, 2, "'%lc'");
@@ -148,27 +151,32 @@ ORBit_trace_value (gconstpointer *val, CORBA_TypeCode tc)
 		if (v == NULL)
 			tprintf("(null)");
 		else {
-			const int max = 64;
+			static int max = -1;
 			const char * v = (*(const char **)*val);
-			char *p, *str = g_strndup (v, max);
+			GString *str;
 			int len = strlen (v);
 
-			if (len > max) {
-				const char ins[5] = " ... ";
-				int i;
-
-				for (i = 0; i < max / 2; ++i)
-					str[i + max / 2] = v[len - max / 2 + i];
-				strncpy(str + (max - sizeof(ins)) / 2, ins, sizeof(ins));
+			if (max < 0) {
+				max = 64;
+				if (g_getenv ("ORBIT2_DEBUG_STRMAX"))
+					max = atoi (g_getenv ("ORBIT2_DEBUG_STRMAX"));
 			}
+			str = g_string_sized_new (max + 8);
+			for (i = 0; i < MIN (max, len); i++) {
+				if (g_ascii_isprint (v[i]) && v[i] != '#')
+					g_string_append_c (str, v[i]);
+				else {
+					g_string_append_c (str, '#');
+					g_string_append_printf (str, "0x%2x", v[i]);
+					g_string_append_c (str, '#');
+				}
+			}
+			if (len > max)
+				g_string_append (str, " ...");
 
-			for (p = str; p && *p; p++)
-				if (!isascii ((int)*p))
-					*p = '#';
+			tprintf ("'%s'", str->str);
 			
-			tprintf ("'%s'", str);
-			
-			g_free (str);
+			g_string_free (str, TRUE);
 		}
 		break;
 	}
@@ -201,8 +209,6 @@ ORBit_trace_value (gconstpointer *val, CORBA_TypeCode tc)
 		break;
 	}
 
-	case CORBA_tk_longlong:
-	case CORBA_tk_ulonglong:
 	case CORBA_tk_longdouble:
 	case CORBA_tk_fixed:
 		tprintf ("wierd");
